import sqlite3
from contextlib import contextmanager
from pathlib import Path

import pytest

from ardilla import Engine, Model, Crud
from ardilla.errors import QueryExecutionError


path = Path(__file__).parent
db = path / "testdb.sqlite"


class User(Model):
    id: int
    name: str
    age: int


@contextmanager
def cleanup():
    try:
        db.unlink(missing_ok=True)
        engine = Engine(db)
        Crud.engine = engine
        crud: Crud[User] = Crud(User)
        engine.setup()
        yield crud
    finally:
        db.unlink(missing_ok=True)


@contextmanager
def query():
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        yield cur
    finally:
        cur.close()
        con.close()


def test_create():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, moni, elena)
        fran = User(id=4, name="fran", age=1)
        crud.save_one(fran)

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 4, "Mistmatch created vs found"

        try:
            crud.insert(id=1, name="john", age=34)
        except QueryExecutionError:
            pass
        else:
            raise Exception("QueryExcecutionError should have been rised")

        u = crud.insert_or_ignore(id=1, name="john", age=34)
        assert u is None, "Record 1 already exists but was somehow re-inserted"

        u = crud.insert_or_ignore(name="john", age=34)
        assert u is not None and u.name == "john", "Bad user was created"


def test_read():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        chris2 = User(id=2, name="chris", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, chris2, elena)

        users = crud.get_all()
        assert len(users) == 3, "Wrong number of read users from get_all"
        users = crud.get_many(name="chris")
        assert len(users) == 2, "wrong number of users from get_many"
        u = crud.get_or_none(id=1)
        assert u == chris, "bad user found"
        u = crud.get_or_none(id=5)
        assert u is None, "bad user not found"
        u, c = crud.get_or_create(name="fran", age=1)
        assert c is True, "Bad created value"
        assert u == User(id=4, name="fran", age=1), "bad created user"
        u, c = crud.get_or_create(id=3)
        assert u == elena, "bad user found"
        assert c is False, "Bad not created value"


def test_update():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        crud.save_many(chris, moni)
        chris.age = 40
        crud.save_one(chris)
        u = crud.get_or_none(**chris.dict())
        assert u is not None, "user not updated"
        chris.age = 45
        moni.age = 50
        crud.save_many(chris, moni)
        u = crud.get_or_none(**moni.dict())
        assert u is not None, "User not updated on many"


def test_delete():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, moni, elena)
        crud.delete_one(moni)
        users = crud.get_all()
        assert len(users) == 2, "Delete one missmatch"
        crud.delete_many(chris, elena)
        users = crud.get_all()
        assert len(users) == 0, "Delete many missmatch"



from __future__ import annotations
import aiosqlite

from ..models import M
from ..engine import Engine


class AsyncEngine(Engine):
    async def connect(self) -> aiosqlite.Connection:
        con = await aiosqlite.connect(self.path)
        con.row_factory = aiosqlite.Row
        return con

    async def __aenter__(self) -> aiosqlite.Connection:
        self.con = await self.connect()
        return self.con

    async def __aexit__(self, *_):
        await self.con.close()

    async def setup(self):
        async with self as con:
            await con.execute("PRAGMA foreign_keys = ON;")
            for table in self.schemas:
                await con.execute(table)
            await con.commit()



class BaseDBException(Exception):
    pass


class ModelIntegrityError(BaseDBException):
    pass


class MissingEngine(BaseDBException):
    pass


class QueryExecutionError(BaseDBException):
    pass



import sqlite3
from sqlite3 import Row
from typing import Literal, Generic, Self

from .engine import Engine
from .abc import CrudABC
from .models import M, Model as BaseModel
from .errors import QueryExecutionError


class Crud(CrudABC, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""

    def _get_or_none_any(self, many: bool, **kws) -> list[M] | M | None:
        """
        private helper to the get_or_none queries.
        if param "many" is true it will return a list of matches else will return only one record
        """
        keys, vals = zip(*kws.items())
        to_match = f" AND ".join(f"{k} = ?" for k in keys)

        limit = "LIMIT 1;" if not many else ";"
        q = f"SELECT rowid, * FROM {self.tablename} WHERE ({to_match}) {limit}"
        with self.engine as con:
            with self.engine.cursor(con) as cur:
                cur.execute(q, vals)
                if many:
                    rows: list[Row] = cur.fetchall()
                    return [self._row2obj(row) for row in rows]
                else:
                    row: Row | None = cur.fetchone()
                    if row:
                        return self._row2obj(row)
        return

    def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        keys, vals = zip(*kws.items())
        placeholders = ", ".join("?" * len(keys))
        cols = ", ".join(keys)

        q = "INSERT OR IGNORE " if ignore else "INSERT "
        q += f"INTO {self.tablename} ({cols}) VALUES ({placeholders})"
        q += " RETURNING *;" if returning else ";"

        with self.engine as con:
            with self.engine.cursor(con) as cur:
                try:
                    cur.execute(q, vals)
                except sqlite3.IntegrityError as e:
                    raise QueryExecutionError(str(e))

                row = cur.fetchone()
                con.commit()
                if returning and row:
                    return self._row2obj(row, cur.lastrowid)

    def get_or_none(self, **kws) -> M | None:
        """Gets an object from a database or None if not found"""
        return self._get_or_none_any(many=False, **kws)

    def insert(self, **kws):
        """
        Inserts a record into the database.
        Returns:
            Model | None: Returns a model only if newly created
        Rises:
            ardilla.error.QueryExecutionError: if there's a conflict when inserting the record
        """
        return self._do_insert(False, True, **kws)

    def insert_or_ignore(self, **kws) -> M | None:
        """inserts a the object of a row or ignores it if it already exists"""
        return self._do_insert(True, True, **kws)

    def get_or_create(self, **kws) -> tuple[M, bool]:
        """Returns object and bool indicated if it was created or not"""
        created = False
        result = self.get_or_none(**kws)
        if not result:
            result = self.insert_or_ignore(**kws)
            created = True
        return result, created

    def get_all(self) -> list[M]:
        """Gets all objects from the database"""
        q = f"SELECT * FROM {self.tablename};"
        with self.engine as con:
            with self.engine.cursor(con) as cur:
                cur.execute(q)
                results = cur.fetchall()
                return [self.Model(**res) for res in results]

    def get_many(self, **kws) -> list[M]:
        """Returns a list of objects that have the given conditions"""
        return self._get_or_none_any(many=True, **kws)

    def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database"""
        cols, vals = zip(*obj.dict().items())
        placeholders = ", ".join("?" * len(cols))

        upsert_query = f"""
        INSERT OR REPLACE INTO {self.tablename} ({', '.join(cols)}) VALUES ({placeholders});
        """
        with self.engine as con:
            con.execute(upsert_query, vals)
            con.commit()
        return True

    def save_many(self, *objs: M) -> Literal[True]:
        """Saves all the given objects to the database"""
        placeholders = ", ".join("?" * len(self.columns))
        q = f'INSERT OR REPLACE INTO {self.tablename} ({", ".join(self.columns)}) VALUES ({placeholders});'
        vals = [tuple(obj.dict().values()) for obj in objs]
        with self.engine as con:
            con.executemany(q, vals)
            con.commit()

        return True

    def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        queries only by the Model id fields (fields suffixed with 'id')
        """
        obj_dict = obj.dict()
        id_cols = tuple([k for k in obj_dict if "id" in k])
        placeholders = ", ".join(f"{k} = ?" for k in id_cols)
        vals = tuple([obj_dict[k] for k in id_cols])
        q = f"""
        DELETE FROM {self.tablename} WHERE ({placeholders});
        """
        with self.engine as con:
            con.execute(q, vals)
            con.commit()

        return True

    def delete_many(self, *objs: M) -> Literal[True]:
        if not objs:
            raise IndexError('param "objs" is empty, pass at least one object')

        prot = objs[0]
        id_cols = tuple([k for k in prot.dict() if "id" in k])
        placeholders = " OR ".join(
            f"({', '.join(f'{k} = ?' for k in id_cols)})" for _ in objs
        )
        vals = tuple(
            [val for obj in objs for key, val in obj.dict().items() if key in id_cols]
        )

        q = f"DELETE FROM {self.tablename} WHERE {placeholders};"

        with self.engine as con:
            con.execute(q, vals)
            con.commit()

        return True



from typing import Literal, Generic, Self

import aiosqlite
from aiosqlite import Row

from .engine import AsyncEngine

from ..errors import QueryExecutionError
from ..models import M
from ..abc import CrudABC


class AsyncCrud(CrudABC, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""

    engine: AsyncEngine

    async def _get_or_none_any(self, many: bool, **kws) -> list[M] | M | None:
        """
        private helper to the get_or_none queries.
        if param "many" is true it will return a list of matches else will return only one record
        """
        keys, vals = zip(*kws.items())
        to_match = f" AND ".join(f"{k} = ?" for k in keys)

        limit = "LIMIT 1;" if not many else ";"
        q = f"SELECT rowid, * FROM {self.tablename} WHERE ({to_match}) {limit}"

        async with self.engine as con:
            async with con.execute(q, vals) as cur:
                if many:
                    rows: list[Row] = await cur.fetchall()
                    return [self._row2obj(row) for row in rows]

                else:
                    row: Row | None = await cur.fetchone()
                    if row:
                        return self._row2obj(row)
        return

    async def get_or_none(self, **kws) -> M | None:
        """Gets an object from a database or None if not found"""
        return await self._get_or_none_any(many=False, **kws)

    async def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        keys, vals = zip(*kws.items())
        placeholders = ", ".join("?" * len(keys))
        cols = ", ".join(keys)

        q = "INSERT OR IGNORE " if ignore else "INSERT "
        q += f"INTO {self.tablename} ({cols}) VALUES ({placeholders})"
        q += " RETURNING *;" if returning else ";"

        async with self.engine as con:
            con = await self.engine.connect()
            cur = None
            try:
                cur = await con.execute(q, vals)
            except aiosqlite.IntegrityError as e:
                raise QueryExecutionError(str(e))
            else:
                row = await cur.fetchone()
                await con.commit()
                if returning and row:
                    return self._row2obj(row, cur.lastrowid)
            finally:
                if cur is not None:
                    await cur.close()
                await con.close()

    async def insert(self, **kws):
        """
        Inserts a record into the database.
        Returns:
            Model | None: Returns a model only if newly created
        Rises:
            ardilla.error.QueryExecutionError: if there's a conflict when inserting the record
        """
        return await self._do_insert(False, True, **kws)

    async def insert_or_ignore(self, **kws) -> M | None:
        """inserts a the object of a row or ignores it if it already exists"""
        return await self._do_insert(True, True, **kws)

    async def get_or_create(self, **kws) -> tuple[M, bool]:
        """Returns object and bool indicated if it was created or not"""
        created = False
        result = await self.get_or_none(**kws)
        if not result:
            result = await self.insert_or_ignore(**kws)
            created = True
        return result, created

    async def get_all(self) -> list[M]:
        """Gets all objects from the database"""
        async with self.engine as con:
            async with con.execute(f"SELECT * FROM {self.tablename};") as cur:
                return [self.Model(**row) for row in await cur.fetchall()]

    async def get_many(self, **kws) -> list[M]:
        """Returns a list of objects that have the given conditions"""
        return await self._get_or_none_any(many=True, **kws)

    async def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database"""
        cols, vals = zip(*obj.dict().items())
        placeholders = ", ".join("?" * len(cols))

        upsert_query = f"""
        INSERT OR REPLACE INTO {self.tablename} ({', '.join(cols)}) VALUES ({placeholders});
        """

        async with self.engine as con:
            await con.execute(upsert_query, vals)
            await con.commit()
        return True

    async def save_many(self, *objs: M) -> Literal[True]:
        """Saves all the given objects to the database"""
        placeholders = ", ".join("?" * len(self.columns))
        q = f'INSERT OR REPLACE INTO {self.tablename} ({", ".join(self.columns)}) VALUES ({placeholders});'
        vals = [tuple(obj.dict().values()) for obj in objs]
        async with self.engine as con:
            await con.executemany(q, vals)
            await con.commit()

        return True

    async def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        queries only by the Model id fields (fields suffixed with 'id')
        """
        obj_dict = obj.dict()
        id_cols = tuple([k for k in obj_dict if "id" in k])
        placeholders = ", ".join(f"{k} = ?" for k in id_cols)
        vals = tuple([obj_dict[k] for k in id_cols])
        q = f"DELETE FROM {self.tablename} WHERE ({placeholders});"
        async with self.engine as con:
            await con.execute(q, vals)
            await con.commit()
        return True

    async def delete_many(self, *objs: M) -> Literal[True]:
        if not objs:
            raise IndexError('param "objs" is empty, pass at least one object')

        prot = objs[0]
        id_cols = tuple([k for k in prot.dict() if "id" in k])
        placeholders = " OR ".join(
            f"({', '.join(f'{k} = ?' for k in id_cols)})" for _ in objs
        )
        vals = tuple(
            [val for obj in objs for key, val in obj.dict().items() if key in id_cols]
        )

        q = f"DELETE FROM {self.tablename} WHERE {placeholders};"
        async with self.engine as con:
            await con.execute(q, vals)
            await con.commit()

        return True



from typing import TypeVar
from pydantic import BaseModel, PrivateAttr

from .schemas import make_schema, FIELD_MAPPING, get_tablename
from .errors import ModelIntegrityError


class Model(BaseModel):
    __rowid__: int | None = PrivateAttr(default=None)
    __tablename__: str  # will default to the lowercase name of the subclass
    __schema__: str  # best effort will be made if it's missing
    # there's no support for constrains or foreign fields

    def __init_subclass__(cls, **kws) -> None:
        for field in cls.__fields__.values():
            if field.type_ not in FIELD_MAPPING:
                raise ModelIntegrityError(
                    f'Field "{field.name}" of model "{cls.__name__}" is of unsupported type "{field.type_}"'
                )

        if not hasattr(cls, "__schema__"):
            cls.__schema__ = make_schema(cls)

        if not hasattr(cls, "__tablename__"):
            tablename = get_tablename(cls)
            setattr(cls, "__tablename__", tablename)

        super().__init_subclass__(**kws)

    def __str__(self) -> str:
        return f"{self!r}"


M = TypeVar("M", bound=Model)



from typing import Generic, Self, Literal
from abc import abstractmethod, ABC
from sqlite3 import Row

from .errors import MissingEngine
from .engine import Engine
from .models import M, Model as BaseModel


class CrudABC(ABC):
    engine: Engine

    def __init__(self, Model: type[M], engine: Engine | None = None) -> None:
        if engine:
            self.engine = engine

        self.Model = Model
        if Model.__schema__:
            self.engine.schemas.add(Model.__schema__)
        self.tablename = Model.__tablename__
        self.columns = tuple(Model.__fields__)

    def __new__(cls, Model: type[M], engine: Engine | None = None) -> Self:
        if not issubclass(Model, BaseModel):
            raise TypeError("Model param has to be a subclass of model")

        cls_engine = getattr(cls, "engine", None)

        if not isinstance(engine, Engine) and not isinstance(cls_engine, Engine):
            raise MissingEngine(
                "You must either set the engine at class level (Crud.engine = Engine(...))"
                "or pass it as an argument."
            )
        return super().__new__(cls)

    def _row2obj(self, row: Row, rowid: int = None) -> M:
        """
        Args:
            row: the sqlite row
            rowid: the rowid of the row.
                If passed it means it comes from an insert function
                meaning the rowid
        """
        [*keys] = self.Model.__fields__
        if rowid is None:
            rowid, *vals = row
        else:
            vals = list(row)
        data = {k: v for k, v in zip(keys, vals)}

        obj = self.Model(**data)
        obj.__rowid__ = rowid
        return obj

    # Create
    @abstractmethod
    def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        pass

    @abstractmethod
    def insert(self, **kws):
        pass

    @abstractmethod
    def insert_or_ignore(self):
        pass

    # Read
    @abstractmethod
    def get_all(self) -> list[M]:
        pass

    @abstractmethod
    def get_many(self, **kws) -> list[M]:
        pass

    @abstractmethod
    def get_or_create(self, **kws) -> tuple[M, bool]:
        pass

    @abstractmethod
    def get_or_none(self, **kws) -> M | None:
        pass

    @abstractmethod
    def _get_or_none_any(self, many: bool, **kws) -> list[M] | M | None:
        pass

    # Update
    @abstractmethod
    def save_one(self, obj: M) -> Literal[True]:
        pass

    @abstractmethod
    def save_many(self, *objs: M) -> Literal[True]:
        pass

    # Delete
    @abstractmethod
    def delete_one(self, obj: M) -> Literal[True]:
        pass

    @abstractmethod
    def delete_many(self, *objs: M) -> Literal[True]:
        pass



from __future__ import annotations
import sqlite3

from .models import M


class ContextCursor:
    def __init__(self, con: sqlite3.Connection):
        self.con = con

    def __enter__(self) -> sqlite3.Cursor:
        self.cur = self.con.cursor()
        return self.cur

    def __exit__(self, *_) -> None:
        self.cur.close()


class Engine:
    def __init__(self, path: str):
        self.path = path
        self.schemas: set[str] = set()

    def __enter__(self) -> sqlite3.Connection:
        con = sqlite3.connect(self.path)
        con.row_factory = sqlite3.Row
        self.con = con
        return con

    def __exit__(self, *_):
        self.con.close()

    def cursor(self, con: sqlite3.Connection) -> ContextCursor:
        return ContextCursor(con)

    def setup(self):
        with self as con:
            con.execute("PRAGMA foreign_keys = ON;")
            for table in self.schemas:
                con.execute(table)
            con.commit()



from datetime import datetime, date
from pydantic import BaseModel


SCHEMA_TEMPLATE = "CREATE TABLE IF NOT EXISTS {tablename} (\n{fields}\n);"


FIELD_MAPPING = {
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bytes: "BLOB",
    bool: "INTEGER",
    date: "DATE",
    datetime: "TIMESTAMP",
}


def get_tablename(model: type[BaseModel]) -> str:
    return getattr(model, "__tablename__", model.__name__.lower())


def get_fields(model: type[BaseModel]) -> str:
    fields = []
    for field in model.__fields__.values():
        if field.type_ not in FIELD_MAPPING:
            raise TypeError(f'Unrecognized sqlite type "{field.type_}"')

        type_ = FIELD_MAPPING[field.type_]

        out = f"    {field.name} {type_}"
        if field.name == "id" and type_ == "INTEGER":
            out += " PRIMARY KEY"
        if field.required and not out.endswith("KEY"):
            out += " NOT NULL"
        if field.default is not None:
            out += f" DEFAULT {field.default!r}"

        fields.append(out)

    return ",\n".join(fields)


def make_schema(Model: type[BaseModel]) -> str:
    return SCHEMA_TEMPLATE.format(
        tablename=get_tablename(Model), fields=get_fields(Model)
    )



from ardilla.models import Model


class User(Model):
    id: int
    name: str


tablename = "user"
schema = """
CREATE TABLE IF NOT EXISTS user (
    id INTEGER PRIMARY KEY,
    name TEXT NOT NULL
);
"""


def test_default_tablename():
    assert (
        User.__tablename__ == "user"
    ), f"wrong default tablename {User.__tablename__} != {tablename}"


def test_defaul_schema():
    assert User.__schema__.strip() == schema.strip()
